/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.konfiguration;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Es wird ein äusseres Strassensegment modelliert.
 *
 * @author Dambach Werke GmbH
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:49:24 $ / ($Author: Sans $)
 */
public class AeusseresStrassenSegment extends StrassenSegment
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Erstellt eine neue Instanz vom Typ <code>AeusseresStrassenSegment</code>
   *
   * @param davObjekt das zuzuordnente Dav-Objekt
   * @param laenge die Länge in m des Strassensegments
   * @param tmcPunkt der TMC-Ort-Code dieses Strassensegments
   * @param tmcRichtung die TMC-Richtung dieses Strassensegments
   * @param strasse Strasse
   */
  public AeusseresStrassenSegment(SystemObject davObjekt, long laenge,
      TmcLocationPunkt tmcPunkt, int tmcRichtung, String strasse)
  {
    super(davObjekt, laenge);
    _tmcPunkt = tmcPunkt;
    _tmcRichtung = tmcRichtung;
    _strasse = strasse;
  }

  /**
   * Der TMC-Ort (TMC Location Code) dieses Straßensegments
   */
  private TmcLocationPunkt _tmcPunkt;

  /**
   * @return den TMC-Ort (TMC Location Code) dieses Straßensegments
   */
  public TmcLocationPunkt getTmcPunkt()
  {
    return _tmcPunkt;
  }

  /**
   * Die TMC-Richtung (TMC Direction) dieses Straßensegments
   */
  private int _tmcRichtung;

  /**
   * Bestimmt die TMC-Richtung (TMC Direction) dieses Straßensegments
   *
   * @return &lt;0 für negative Richtung und &gt;0 für positive Richtung,
   * 0 wenn Richtung nicht definiert
   */
  public int getTmcRichtung()
  {
    return _tmcRichtung;
  }

  /**
   * Die Straße, aud der dieses Straßensegment liegt
   */
  private String _strasse;

  /**
   * @return ie Straße, aud der dieses Straßensegment liegt
   */
  public String getStrasse()
  {
    return _strasse;
  }

  /**
   * Prüft, ob 2 äußere Straßensegmente auf einer Straße liegen
   *
   * @param segment das mit diesem Segment zu prüfende äußere Segment
   * @return true, wenn beide Segmente auf einer Straße leigen, false sonst
   */
  public boolean liegtAufSelberStrasse(AeusseresStrassenSegment segment)
  {
    return (segment != null &&
            getStrasse().equals(segment.getStrasse()) &&
            (getTmcRichtung() == segment.getTmcRichtung()));
  }

  /**
   * Der Strassenknoten, an dem dieses Strassensegment beginnt
   */
  private StrassenKnoten _vonKnoten;

  /**
   * @return den Strassenknoten, an dem dieses Strassensegment beginnt
   */
  public StrassenKnoten getVonKnoten()
  {
    return _vonKnoten;
  }

  /**
   * Setzt den Startknoten für dieses Strassensegment
   *
   * @param der Strassenknoten, an dem dieses Strassensegment beginnt
   */
  void setVonKnoten(StrassenKnoten vonKnoten)
  {
    _vonKnoten = vonKnoten;
  }

  /**
   * Der Strassenknoten, an dem dieses Strassensegment endet
   */
  private StrassenKnoten _nachKnoten;

  /**
   * @return den Strassenknoten, an dem dieses Strassensegment endet
   */
  public StrassenKnoten getNachKnoten()
  {
    return _nachKnoten;
  }

  /**
   * Setzt den Endknoten für dieses Strassensegment
   *
   * @param der Strassenknoten, an dem dieses Strassensegment endet
   */
  void setNachKnoten(StrassenKnoten nachKnoten)
  {
    _nachKnoten = nachKnoten;
  }

  /**
   * Methode prüft das Straßensegmenten in einem Autobahnkreuz oder Autobahndreieck beginnt
   *
   * @return Autobahnkreuz, Autobahndreick: true, sonst false
   */
  @Override
  public boolean beginntInAutobahnKreuzOderDreieck ()
  {
    return (_vonKnoten != null &&
      (_vonKnoten.isAutobahnKreuz() || _vonKnoten.isAutobahnDreieck()));
  }

  /**
   * Methode prüft das Straßensegmenten in einem Autobahnkreuz oder Autobahndreieck endet
   *
   * @return Autobahnkreuz, Autobahndreick: true, sonst false
   */
  @Override
  public boolean endetInAutobahnKreuzOderDreieck ()
  {
    return (_nachKnoten != null &&
        (_nachKnoten.isAutobahnKreuz() || _nachKnoten.isAutobahnDreieck()));
  }

  /**
   * Methode prüft ob das Straßensegmenten am Autobahnanfang beginnt
   *
   * @return Autobahnanfang: true, sonst false
   */
  @Override
  public boolean beginntAmAutobahnAnfang ()
  {
    return (_vonKnoten != null && _vonKnoten.isAutobahnAnfang());
  }

  /**
   * Methode prüft ob das Straßensegmenten am Autobahnende endet
   *  
   * @return Autobahnende: true, sonst false 
   */
  @Override
  public boolean endetAmAutobahnEnde ()
  {
    return (_nachKnoten != null && _nachKnoten.isAutobahnEnde());
  }
}
