/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.konfiguration;

import java.util.Collection;
import java.util.Vector;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Es wird ein inneres Strassensegment modelliert.
 * 
 * @author Dambach Werke GmbH
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:49:24 $ / ($Author: Sans $)
 */
public class InneresStrassenSegment extends StrassenSegment 
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Das Vorgänger-Segment
   */
  private AeusseresStrassenSegment _vonSegment;

  /**
   * @return das Vorgänger-Segment
   */
  public AeusseresStrassenSegment getVonSegment()
  {
    return _vonSegment;
  }

  /**
   * Das Nachfolger-Segment
   */
  private AeusseresStrassenSegment _nachSegment;

  /**
   * @return das Nachfolger-Segment
   */
  public AeusseresStrassenSegment getNachSegment()
  {
    return _nachSegment;
  }

  /**
   * Erstellt eine neue Instanz vom Typ <code>InneresStrassenSegment</code>
   *
   * @param davObjekt das zuzuordnente Dav-Objekt
   * @param laenge die Länge in m des Strassensegments
   */
  public InneresStrassenSegment(SystemObject davObjekt, long laenge)
  {
    super(davObjekt, laenge);
  }

  @Override
  /**
   * Fügt diesem Strassensegment einen Nachfolger hinzu
   */
  void addNachfolger(StrassenSegment segment)
  {
    _nachSegment = (AeusseresStrassenSegment)segment;
    _nachfolger.clear();
    super.addNachfolger(segment);
  }

  @Override
  /**
   * Fügt diesem Strassensegment einen Vorgänger hinzu
   */
   void addVorgaenger(StrassenSegment segment)
  {
    _vonSegment = (AeusseresStrassenSegment)segment;
    _vorgaenger.clear();
    super.addVorgaenger(segment);
  }

  @Override
  /**
   * Fügt diesem Strassensegment einen Nachfolger auf derselben Strasse hinzu
   */
  void addNachfolgerAufStrasse(StrassenSegment segment)
  {
    _nachfolgerAufStrasse.clear();
    super.addNachfolgerAufStrasse(segment);
  }

  @Override
  /**
   * Fügt diesem Strassensegment einen Vorgänger auf derselben Strasse hinzu
   */
   void addVorgaengerAufStrasse(StrassenSegment segment)
  {
    _vorgaengerAufStrasse.clear();
    super.addVorgaengerAufStrasse(segment);
  }

  /**
   * Methode prüft das Straßensegmenten in einem Autobahnkreuz oder Autobahndreieck liegt
   *
   * @return Autobahnkreuz, Autobahndreick: true, sonst false
   */
  @Override
  public boolean beginntInAutobahnKreuzOderDreieck ()
  {
    // Ein inneres Straßensegment hat keinen direkten Bezug zum Straßenknoten, daher erfolgt der
    // Zugriff über die durch das innere Straßensegment verbundenen äußeren Straßenknoten

    if (_nachSegment != null)
    {
      return _nachSegment.beginntInAutobahnKreuzOderDreieck();
    }
    if (_vonSegment != null)
    {
      return _vonSegment.endetInAutobahnKreuzOderDreieck();
    }
    return false;
  }

  /**
   * Methode prüft das Straßensegmenten in einem Autobahnkreuz oder Autobahndreieck endet
   *
   * @return Autobahnkreuz, Autobahndreick: true, sonst false
   */
  @Override
  public boolean endetInAutobahnKreuzOderDreieck ()
  {
    // Da ein inneres Straßensegment innerhalb eines Straßenknotens liegt, braucht nicht zwischen
    // beginnt und endet unterschieden werden.

    return beginntInAutobahnKreuzOderDreieck();
  }

  /**
   * Methode prüft ob der Straßenknoten in dem das Straßensegment liegt ein Autobahnanfang ist
   *
   * @return Autobahnanfang: true, sonst false
   */
  @Override
  public boolean beginntAmAutobahnAnfang ()
  {
    // Ein inneres Straßensegment hat keinen direkten Bezug zum Straßenknoten, daher erfolgt der
    // Zugriff über die durch das innere Straßensegment verbundenen äußeren Straßenknoten

    return (_nachSegment != null && _nachSegment.beginntAmAutobahnAnfang());
  }

  /**
   * Methode prüft ob der Straßenknoten in dem das Straßensegment liegt ein Autobahnende ist
   *
   * @return Autobahnende true, sonst false
   */
  @Override
  public boolean endetAmAutobahnEnde ()
  {
    // Ein inneres Straßensegment hat keinen direkten Bezug zum Straßenknoten, daher erfolgt der
    // Zugriff über die durch das innere Straßensegment verbundenen äußeren Straßenknoten
    
    return (_vonSegment != null &&  _vonSegment.endetAmAutobahnEnde());
  }

}
