/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
 
package de.bsvrz.ste.rdstmc.meldungserstellung;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Abtrakte Klasse zur Verwaltung von dynamischen DAV-Objekten.
 * Diese abstrakte Klasse stellt Funktionalität zur Vefügung um generische Instanzen
 * für dynamische DAV-Objekte, die in einer dynamischen Menge enthalten sind, zu verwalten<br>.
 * Die verwalteten Instanzen werden in einer (nich modifizierbaren) Map mit der PID
 * des dynamischen Objekts als Schlüssel zur Verfügung gestellt. Diese Map wird
 * entsprechend den Änderungen in der dynamischen Objektmenge und den Löschungen
 * (Invalidierungen) der dynamischen Objekte aktualisiert.<br>
 * Zur Implementierung einer effektiven Verwaltung genügt die Implementierung
 * der Methode <code>erzeugeObjekt</code> und aufrufen der Methode <code>initialisiere</code>
 * z. B. im Konstruktor der beerbenden Klasse.
 *
 * @author Dambach Werke GmbH
 * @version $Revision: 1.2 $ / $Date: 2009/04/03 08:58:04 $ / ($Author: Sans $)
 */
public abstract class AbstraktVewDynObj<T> implements MutableSetChangeListener,
  InvalidationListener
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Die verwalteten Instanzen vom Typ <code>T</code>
   */
  private Map<String, T> _verwalteteObjekte = new HashMap<String, T>();

  /**
   * Read-only Zugriff auf die verwalteten Instanzen vom Typ <code>T</code>
   *
   * @return Map (nicht modifizierbar) der verwalteten Objekte, Schlüssel ist die PID
   * des dynamischen Objekts
   */
  public Map<String, T> getVerwalteteObjekte()
  {
    return Collections.unmodifiableMap(_verwalteteObjekte);
  }

  /**
   * Erzeuge eine neue Instanz vom verwalteten Objekttyp <code>T</code>
   *
   * @param sysObj das dynamische DAV-Objekt, zu dem die Instanz
   * vom Typ <code>T</code> erzeugt wird
   *
   * @return das erzeugte Objekt vom Typ <code>T</code>
   */
  protected abstract T erzeugeObjekt(SystemObject sysObj);

  /**
   * Muss aufgerufen werden, um die Verwaltung für die aktuell in  der dynamischen Menge
   * enthaltenen Objekten zu initialisieren - z. B. als Bestandteil des Konstruktors.
   *
   * @param konfigurationsObjekt, das Besitzer der dynamischen Menge ist
   * @param mengenBezeichner Bezeichner (Name) der Menge
   */
  protected void initialisiere(ConfigurationObject konfigurationsObjekt, String mengenBezeichner)
  {
    MutableSet menge = konfigurationsObjekt.getMutableSet(mengenBezeichner);
    if (menge != null)
    {
      int anzahl = 0;

      for (SystemObject obj: menge.getElements())
      {
        _debug.finer("Lese Objekt: " + obj + "\t" + obj.getType() + "\t" + obj.isValid());
        if (obj.isValid())
        {
          if (initialisiereObjekt(obj) != null)
          {
            anzahl++;
          }
        }
      }

      _debug.config("Verwalte " + anzahl + " Element(e) aus der dynamischen Menge \"" +
          mengenBezeichner + "\" von " + konfigurationsObjekt);

      menge.addChangeListener(this);
    }
    else
    {
      throw new IllegalArgumentException("Dynamische Menge \"" + mengenBezeichner +
          "\" nicht vorhanden in " + konfigurationsObjekt);
    }
  }

  /**
   * Wird von der Verwaltung aufgerufen und führt Initialisierungen durch, wenn ein
   * dynamisches Objekt erzeugt oder der dynamischen Menge hinzugefügt wird
   *
   * @param sysObj das dynamische Object, das hinzugefügt wird
   * @return die Instanz vom Typ <code>T</code>, für die die Verwaltung initialisiert
   * wurde - für diese Instanz können jetzt ggf. weitere Resourcen belegt werden.
   * Wenn die Initialisierung nicht ergolgreich war wird <code>null</code> zurückgegenen
   */
  protected T initialisiereObjekt(SystemObject sysObj)
  {
    if (_verwalteteObjekte.get(sysObj.getPid()) == null)
    {
      T obj = erzeugeObjekt(sysObj);

      if (obj != null)
      {
        synchronized(this)
        {
          _verwalteteObjekte.put(sysObj.getPid(), obj);
          if (sysObj instanceof DynamicObject)
          {
            ((DynamicObject)sysObj).addListenerForInvalidation(this);
          }
        }
      }

      return obj;
    }
    else
    {
      _debug.warning("Objekt mit PID " + sysObj.getPid() +
          " wird bereits verwaltet");
      return null;
    }
  }

  /**
   * Wird von der Verwaltung aufgerufen und führt Bereinigungen durch, wenn ein
   * dynamisches Ovjekt gelöscht oder aus der dynamischen Menge entfernt wird
   *
   * @param sysObj das dynamische Object, das entfernt wird
   * @return die Instanz vom Typ <code>T</code>, für das die Verwaltung bereinigt
   * wurde - für diese Instanz können jetzt ggf. weiter Resourcen freigegeben werden.
   * Wenn die Bereinigung nicht erfolgreich war wird <code>null</code> zurückgegenen
   */
  protected T bereinigeObjekt(SystemObject sysObj)
  {
    T obj = null;

    if (sysObj instanceof DynamicObject)
    {
      ((DynamicObject)sysObj).removeListenerForInvalidation(this);
    }

    synchronized(this)
    {
      if (_verwalteteObjekte.get(sysObj.getPid()) != null)
      {
        obj = _verwalteteObjekte.remove(sysObj.getPid());
      }
    }

    return obj;
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet, de.bsvrz.dav.daf.main.config.SystemObject[], de.bsvrz.dav.daf.main.config.SystemObject[])
   */
  public void update(MutableSet set, SystemObject[] addedObjects,
      SystemObject[] removedObjects)
  {
    _debug.finer("MutableSetChangeListener.update() von \"" + set.getName() +"\"");

    if (addedObjects.length > 0)
    {
      _debug.config("Dynamische Menge \"" + set.getName() + "\" um " +
          addedObjects.length + " Objekte vergrössert");

      for(SystemObject obj: addedObjects)
      {
        initialisiereObjekt(obj);
      }
    }

    if (removedObjects.length > 0)
    {
      _debug.config("Dynamische Menge \"" + set.getName() + "\" um " +
          removedObjects.length + " Objekte verkleinert");

      for(SystemObject obj: removedObjects)
      {
        bereinigeObjekt(obj);
      }
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.config.InvalidationListener#invalidObject(de.bsvrz.dav.daf.main.config.DynamicObject)
   */
  public void invalidObject(DynamicObject dynamicObject)
  {
    _debug.config("Objekt " + dynamicObject +
        " gelöscht/invalidiert => wird nicht weiter verwaltet");
    
    bereinigeObjekt(dynamicObject);
  }
    
}
