/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.meldungserstellung;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.rdstmc.allgemein.RDSEreignis;
import de.bsvrz.ste.rdstmc.konfiguration.Konfiguration;
import de.bsvrz.ste.rdstmc.konfiguration.TmcLocationPunkt;
import de.bsvrz.sys.funclib.dambach.parameter.ParameterEmpfaenger;
import de.bsvrz.sys.funclib.debug.Debug;

public class VerwaltungMeldungen implements InvalidationListener, MeldungenGlobal
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  private ClientDavInterface _connection;

  public ClientDavInterface getConnection()
  {
    return _connection;
  }

  private ConfigurationObject _konfigurationsObjekt;

  public ConfigurationObject getKonfigurationsObjekt()
  {
    return _konfigurationsObjekt;
  }

  private Konfiguration _konfiguration;

  /**
   * Datenbeschreibung "generiert"
   */
  private DataDescription _datenBeschreibung;

  public DataDescription getDatenBeschreibung()
  {
    return _datenBeschreibung;
  }

  private MeldungenGlobal _meldungenGlobal;

  /***************************************************************************/
  /* Implementierung Interface MeldungenGlobal                               */
  /***************************************************************************/

  /**
   * @return den Konfigurationsbereich, in dem dynamische Meldungsobjekte abgelegt werden
   */
  public ConfigurationArea getKonfigurationsbereich()
  {
    return _meldungenGlobal.getKonfigurationsbereich();
  }

  /**
   * @return die LocationMethode, nach der Meldungsdatensätze kodiert werden
   */
  public LocationMethode getLocationMethode()
  {
    return _meldungenGlobal.getLocationMethode();
  }

  /**
   * @return die Methode, nach der beim Programmstart aktuelle Zustände publiziert werden
   */
  public PublikationInitial getPublikationInitial()
  {
    return _meldungenGlobal.getPublikationInitial();
  }

  /**
   * @return die (Text-) Id der Organisationseinheit, die einer generierten Meldung zugeordnet sein soll
   * - i. d. R. LMS
   */
  public String getIdOrganisation()
  {
    return _meldungenGlobal.getIdOrganisation();
  }

  /**
   * @return die (numerische) Id der Organisationseinheit, die einer generierten Meldung zugeordnet sein soll
   * - i. d. R. LMS
   */
  public String getIdNummer()
  {
    return _meldungenGlobal.getIdNummer();
  }

  /**
   * @return Zeitspanne in min., wielange eine Meldung gültig (nicht abgelaufen)
   * sein soll, nachdem sie aktualisiert wurde
   */
  public int getAblaufOffset()
  {
    return _meldungenGlobal.getAblaufOffset();
  }

  /***************************************************************************/
  /* Konstanten für die MEldungsverwaltung                                   */
  /***************************************************************************/

  public final static String MENGE = "RDSMeldungen";

  public final static String PID_PREFIX = "rdsMeldung";

  public final static String NAME_PREFIX = "RDSMeldung";

  public VerwaltungMeldungen(ClientDavInterface connection, ConfigurationObject konfigurationsObjekt,
      Konfiguration konfiguration, MeldungenGlobal meldungenGlobal)
  {
    _connection = connection;
    _konfigurationsObjekt = konfigurationsObjekt;
    _konfiguration = konfiguration;
    _meldungenGlobal = meldungenGlobal;

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.rdsMeldung");
    Aspect asp = getConnection().getDataModel().getAspect("asp.rdsGeneriert");
    _datenBeschreibung = new DataDescription(atg, asp);
  }

  private Map<String, Meldung> _meldungen = new LinkedHashMap<String, Meldung>();

  public Meldung getMeldung(String schluessel)
  {
    synchronized(this)
    {
      return _meldungen.get(schluessel);
    }
  }

  private Ereignis bestimmeEreignis(ResultData resultData)
  {
    Data data = resultData.getData();
    List<TmcLocationPunkt> tmcLocationen = new LinkedList<TmcLocationPunkt>();

    switch (getLocationMethode())
    {
      case LISTE:
      {
        List<String> locationCodes = getAlleTmcLocationen(data);
        for (String code: locationCodes)
        {
          tmcLocationen.add(_konfiguration.getTmcPunkteMitLocationCode().get(code));
        }
        break;
      }
      case EXTEND:
      {
        TmcLocationPunkt tmcPunkt = _konfiguration.getTmcPunkteMitLocationCode().get(getPrimaereTmcLocation(data));
        int extend = getLocationExtend(data);
        while (tmcLocationen.size() <= extend)
        {
          tmcLocationen.add(tmcPunkt);
          tmcPunkt = tmcPunkt.getVorgaenger();
        }
        break;
      }
      default:
      {
        throw new UnsupportedOperationException("Location-Methode mit Id \"" +
            getLocationMethode().id + "\" nicht unterstützt");
      }
    }

    int tmcRichtung = getTmcRichtung(data);
    String ereignisCode = getEreignisCode(data);
    RDSEreignis rdsEreignis = RDSBeschreibung.getRDSEreignisZuCode(ereignisCode);
    RDSQuantitaetLEN rdsQuantitaetLEN = null;

    String wert = getRDSQuantitaetLENWert(data);
    if (wert != null && !wert.equals(""))
    {
      rdsQuantitaetLEN = new RDSQuantitaetLEN(Short.parseShort(wert));
    }

    Ereignis ereignis = new Ereignis(Ereignis.Art.INITIAL, rdsEreignis, tmcLocationen, tmcRichtung, ereignisCode);
    ereignis.setRDSQuantitaetLEN(rdsQuantitaetLEN);

    return ereignis;
  }

  public Map<String, Ereignis> initialisiereVorhandeneMeldungen()
  {
    MutableSet menge = getKonfigurationsObjekt().getMutableSet(MENGE);
    if (menge != null)
    {
      synchronized(this)
      {
        int anzahl = 0;
        Map<String, Ereignis> result = new LinkedHashMap<String, Ereignis>();
        _meldungen.clear();

        for (SystemObject objekt: menge.getElements())
        {
          if (objekt.isValid())
          {
            _debug.finer("Bearbeite dynamisches Objekt " + objekt.getPid());

            ResultData datenGeneriert = new ParameterEmpfaenger(getConnection(), objekt,
                "atg.rdsMeldungGeneriert", _datenBeschreibung).getData();

            if (datenGeneriert != null && datenGeneriert.hasData())
            {
              Meldung meldung = null;
              String schluessel = null;
              DynamicObject dynObj = null;
              try
              {
                Ereignis ereignis = bestimmeEreignis(datenGeneriert);
                schluessel = ereignis.getSchluessel();
                meldung = new Meldung(this, objekt, datenGeneriert);

                if (!istAufhebung(datenGeneriert.getData()))
                {
                  // falls Ereignis keine Aufhebung ist
                  _meldungen.put(schluessel, meldung);
                  result.put(schluessel, ereignis);
                }
                if (objekt instanceof DynamicObject)
                {
                  dynObj = (DynamicObject)objekt;
                  dynObj.addListenerForInvalidation(this);
                }
                anzahl += 1;
                _debug.fine("Dynamisches Objekt " + objekt.getPid() + " fehlerfrei initialisiert");

                switch(getPublikationInitial())
                {
                  case AENDERUNGEN:
                    // tue nichts
                    break;
                  case ALLES:
                    meldung.send(meldung.getDatenGeneriert());
                    break;
                }
              }
              catch (RuntimeException e)
              {
                if (meldung != null)
                {
                  if (dynObj != null)
                  {
                    dynObj.removeListenerForInvalidation(this);
                  }
                  if (schluessel != null)
                  {
                    _meldungen.remove(schluessel);
                  }
                  meldung.dispose();
                }
                _debug.warning("Initialisierung fehlgeschlagen für dynamisches Objekt " + objekt.getPid() +
                    ": Objekt wird nicht verwendet", e);
              }
            }
            else
            {
              _debug.fine("Dynamisches Objekt " + objekt.getPid() +
                  " wurde nicht initialisiert, da kein Datensatz unter dem Aspekt \"RDSGenereiert\" vorhanden ist");
            }
          }
        }
        _debug.config(anzahl + " vorhandene RDS-Meldungen werden verwaltet");
        return result;
      }
    }
    else
    {
      throw new IllegalArgumentException("Dynamische Menge \"" + MENGE +
          "\" nicht vorhanden in " + getKonfigurationsObjekt());
    }
  }

  public boolean erzeugeMeldung(String schluessel)
  {
    DynamicObjectType typ = (DynamicObjectType)getConnection().getDataModel().getType("typ.rdsMeldung");
    DynamicObject davObjekt = null;
    short ext = 1;

    while(ext < Short.MAX_VALUE && getConnection().getDataModel().getObject(PID_PREFIX + ext) != null)
    {
      ext += 1;
    }

    try
    {
      davObjekt = getKonfigurationsbereich().createDynamicObject(typ, PID_PREFIX + ext, NAME_PREFIX + ext, null);
      getKonfigurationsObjekt().getMutableSet(MENGE).add(davObjekt);
    }
    catch(ConfigurationChangeException e)
    {
      if (davObjekt != null)
      {
        try
        {
          davObjekt.invalidate();
        }
        catch(ConfigurationChangeException ex)
        {
          // man kann nichts mehr tun
        }
      }
      _debug.error("Fehler bei Erzeugen eines dynamisches Objekts vom Typ " + typ.getPid(), e);
    }

    if (davObjekt != null)
    {
      synchronized(this)
      {
        _meldungen.put(schluessel, new Meldung(this, davObjekt));
        _debug.finer("Neue RDS-Meldung " + davObjekt.getName() + " ohne Fehler erzeugt");
        return true;
      }
    }
    else
    {
      return false;
    }
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.config.InvalidationListener#invalidObject(de.bsvrz.dav.daf.main.config.DynamicObject)
   */
  public void invalidObject(DynamicObject dynObjekt)
  {
    _debug.config("Objekt " + dynObjekt +
        " gelöscht/invalidiert => wird nicht weiter verwaltet");
    synchronized(this)
    {
      Iterator<String> it = _meldungen.keySet().iterator();

      while (it.hasNext())
      {
        String schluessel = it.next();
        Meldung meldung = _meldungen.get(schluessel);
        if (meldung.getDavObjekt() == dynObjekt)
        {
          meldung.dispose();
          dynObjekt.removeListenerForInvalidation(this);
          it.remove();
        }
      }
    }
  }

  private static String getEreignisCode(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataEreignis = dataVerkehr.getItem("Ereignis");
    Array dataEreignisDaten = dataEreignis.getArray("EreignisDaten");
    Data ereignisDaten = dataEreignisDaten.getItem(0);
    return "" + ereignisDaten.getItem("EreignisCode").asUnscaledValue().intValue();
  }

  private static String getPrimaereTmcLocation(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
    Array dataRDSLocation = dataLocationDaten.getArray("RDSLocation");
    Data locationDaten = dataRDSLocation.getItem(0);
    return locationDaten.getItem("LocationCode").asTextValue().getText();
  }

  private static List<String> getAlleTmcLocationen(Data data)
  {
    List<String> result = new LinkedList<String>();

    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
    Array dataRDSLocation = dataLocationDaten.getArray("RDSLocation");
    for (int i = 0; i < dataRDSLocation.getLength(); i += 1)
    {
      Data locationDaten = dataRDSLocation.getItem(i);
      result.add(locationDaten.getItem("LocationCode").asTextValue().getText());
    }
    return result;
  }

  private static int getTmcRichtung(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
    return dataLocationDaten.getItem("TMCRichtung").asUnscaledValue().intValue();
  }

  private static int getLocationExtend(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
    return dataLocationDaten.getItem("LocationExtent").asUnscaledValue().intValue();
  }

  private static String getRDSQuantitaetLENWert(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
    Data dataEreignis = dataVerkehr.getItem("Ereignis");
    Array dataEreignisDaten = dataEreignis.getArray("EreignisDaten");
    if (dataEreignisDaten.getLength() > 0)
    {
      Data ereignisDaten = dataEreignisDaten.getItem(0);
      Array dataEreignisQuantitaet = ereignisDaten.getArray("EreignisQuantität");
      if (dataEreignisQuantitaet.getLength() > 0)
      {
        Data quantitaetDatenDaten = dataEreignisQuantitaet.getItem(0);
        return quantitaetDatenDaten.getItem("QuantitätsWert").asTextValue().getText();
      }
    }
    return null;
  }

  private static boolean istGeneriert(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
    return dataVerwaltung.getItem("Quelle").asTextValue().getText().equals("automatisch");
  }

  private static boolean istAufhebung(Data data)
  {
    Data dataVersion = data.getItem("Version");
    Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
    String status = dataVerwaltung.getItem("Status").asTextValue().getText();
    return status.equals("Aufhebung") || status.equals("Löschung");
  }

}
