/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.verkehrsanalyse;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.rdstmc.allgemein.StoerfallZustand;
import de.bsvrz.ste.rdstmc.konfiguration.DavProxyObjekt;
import de.bsvrz.ste.rdstmc.konfiguration.StrassenSegment;
import de.bsvrz.ste.rdstmc.parameter.RDSParameterDaten;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modelliert einen Störfall in Bezug auf DAV-Kommunikation
 *
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.2 $ / $Date: 2008/11/10 13:57:55 $ / ($Author: Sans $)
 */
public abstract class Stoerfall implements ClientReceiverInterface
{

  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  private static GlaettungsThread _glaettungsThread = new GlaettungsThread();

  /**
   * Korrespondierendes DAV-Systemobjekt
   */
  private DavProxyObjekt _davProxy;

  /**
   * StoerfallSituation, zu der dieser Störfall gehört
   */
  private StoerfallSituation _stoerfallSituation;

  /**
   * @return die StoerfallSituation, zu der dieser Störfall gehört
   */
  public StoerfallSituation getStoerfallSituation()
  {
    return _stoerfallSituation;
  }

  /**
   * Die Attributgruppe für Störfall-Zustände
   */
  private AttributeGroup _atgStoerfallZustand;

  /**
   * Sammlung aller <code>StoerfallListener</code>
   */
  private Collection<StoerfallListener> _stoerfallListener = new ArrayList<StoerfallListener>();

  public void addStoerfallListener(StoerfallListener listener)
  {
    synchronized(_stoerfallListener)
    {
      _stoerfallListener.add(listener);
    }
  }

  public void renmoveStoerfallListener(StoerfallListener listener)
  {
    synchronized(_stoerfallListener)
    {
      _stoerfallListener.remove(listener);
    }
  }

  protected void fireStoerfallListener()
  {
    Daten dominator = null;
    int prio = -1;
    Collection<Aspect>  alleAspekte = new HashSet<Aspect>(_atgStoerfallZustand.getAspects());

    // Bestimme Datensatz mit höchster Priorität unter allen aktuell
    // parametrierten Verfahren
    for (Aspect aspekt: getParameter().getRDSSIVerfahren())
    {
      Daten daten = _aktuelleDaten.get(aspekt.getPid());
      if (daten != null)
      {
        int prioDaten = getParameter().getRdsPrioritaetenMatrix(daten.zustand, daten.guete);
        if (prioDaten > prio)
        {
          prio = prioDaten;
          dominator = daten;
        }
      }
      alleAspekte.remove(aspekt);
    }

    // Bereinige die derzeit nicht parametrierten Verfahren
    for (Aspect aspekt: alleAspekte)
    {
      _aktuelleDaten.remove(aspekt);
    }

    StoerfallDaten daten = dominator == null ? null :
      new StoerfallDaten(this, dominator.zustand, prio, getStartOffset(), getEndOffset());

    synchronized(_stoerfallListener)
    {
      for (StoerfallListener listener: _stoerfallListener)
      {
        listener.neuerStoerfall(this, daten);
      }
    }
  }

  ClientDavInterface getConnection()
  {
    return _stoerfallSituation.getConnection();
  }

  RDSParameterDaten getParameter()
  {
    return _stoerfallSituation.getParameter();
  }

  /**
   * Konstruiert eine Instanz vom Typ <code>Stoerfall</code>
   *
   * @param davObjekt das korrespondierende DAV-Systemobjekt
   * @param stoerfallSituation StoerfallSituation
   */
  public Stoerfall(DavProxyObjekt davObjekt, StoerfallSituation stoerfallSituation)
  {
    _davProxy = davObjekt;
    _stoerfallSituation = stoerfallSituation;

    _atgStoerfallZustand = getConnection().getDataModel().getAttributeGroup("atg.störfallZustand");

    for (Aspect aspekt: _atgStoerfallZustand.getAspects())
    {
      getConnection().subscribeReceiver(this, getDavObjekt(), new DataDescription(_atgStoerfallZustand,
          aspekt), ReceiveOptions.normal(), ReceiverRole.receiver());
      _debug.finer(_atgStoerfallZustand.getPid() + "/" + aspekt.getPid() +
          " zum Empfang angemeldet für " + _davProxy);
    }
  }

  /**
   * @return das Strassensegment, auf dem dieser Störfall liegt
   */
  StrassenSegment getSegment()
  {
    return getStoerfallSituation().getSegment();
  }

  /**
   * @return den Startoffset in m auf dem Straßensegment dieses Stöfalls
   */
  abstract long getStartOffset();

  /**
   * @return den Endoffset in m auf dem Straßensegment dieses Stöfalls
   */
  abstract long getEndOffset();

  /**
   * Container für die aktuell gültigen Störfall-Daten per Verfahren
   */
  private HashMap<String, Daten> _aktuelleDaten = new HashMap<String, Daten>();


  /**
   * @return das diesem Störfall zugeordnete Objekt vom Typ <code>DavProxyObjekt</code>
   */
  protected DavProxyObjekt getObjekt()
  {
    return _davProxy;
  }

  /**
   * @return das diesem Störfall zugeordnete DAV-Objekt
   */
  public SystemObject getDavObjekt()
  {
    return getObjekt().getDavObjekt();
  }


  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] resultData)
  {
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();

      if (_atgStoerfallZustand.equals(atg) && atg.getAspects().contains(asp))
      {
        StoerfallZustand zustand = null;
        float guete = 0.0f;

        if (data.hasData())
        {
          try
          {
            int id = data.getData().getItem("Situation").asUnscaledValue().intValue();
            zustand = StoerfallZustand.get(id);

            if (zustand != null)
            {
              if (zustand == StoerfallZustand.STOERUNG)
              {
                _debug.warning("update() für " + _davProxy + "/"  + asp.getPid() +
                    ": Zustand \""  + zustand + "\" empfangen");
                // Zustand "Störung" wird behandelt als "keine Aussage" mit Güte 1.0
                zustand = StoerfallZustand.KEINE_AUSSAGE;
                guete = 1.0f;
              }
              else
              {
                Data datenSatz = data.getData().getItem("Güte");

                guete = datenSatz.getItem("Index").asScaledValue().floatValue();

                if (guete < 0.0 || guete > 1.0)
                {
                  _debug.warning("update() für " + _davProxy + "/"  + asp.getPid() +
                      ": Ungültger Wert für Güte-Index " +
                      datenSatz.getItem("Index").asTextValue().getText());
                  // Ungültiger Gütewert wird behandelt als "keine Aussage" mit Güte 1.0
  //                zustand = StoerfallZustand.KEINE_AUSSAGE;
  //                guete = 1.0f;

                  // Ungültiger Gütewert wird ignoriert
                  return;
                }
              }
            }
            else
            {
              _debug.warning("update() für " + _davProxy + "/"  + asp.getPid() +
                  ": Zustand mit Id " + id + " konnte nicht aufgelöst werden");
              // Ungültiger Zustandswert wird behandelt als "keine Aussage" mit Güte 1.0
  //            zustand = StoerfallZustand.KEINE_AUSSAGE;
  //            guete = 1.0f;

              // Ungültiger Zustand wird ignoriert
              return;
            }
          }
          catch(IllegalStateException e)
          {
            _debug.warning("Fehler", e);
          }
        }
        else
        {
          // "Keine Daten" (auch "keine Quelle") wird behandelt als "keine Aussage" mit Güte 1.0
          zustand = StoerfallZustand.KEINE_AUSSAGE;
          guete = 1.0f;
        }

         _glaettungsThread.put(new Daten(asp, zustand, guete, getConnection().getTime()));
         _debug.finer("Update Störfall-Zustand für " + getObjekt() + "/" + asp.getPid() + ": Zustand \"" +
             zustand + "\" mit Güte " + guete);

      }
    }
  }

  boolean uebernehmeDaten(Daten soll)
  {
    boolean result = true;
    String verfahren = soll.aspekt.getPid();

    synchronized(getStoerfallSituation().getVerkehrsanalyse().mutexStoerfall)
    {
      if (getParameter().getRDSSIVerfahren().contains(soll.aspekt))
      {
        Daten ist = _aktuelleDaten.get(verfahren);
        if (ist == null) // Sofortübernahme, da noch keine Daten für dieses Verfahren übernommen
        {
          _aktuelleDaten.put(verfahren, soll);
          if (soll.zustand == StoerfallZustand.KEINE_AUSSAGE)
          {
            _debug.finest("Übernahme Störfall-Zustand für " + _davProxy + "/" +
                verfahren + ": " + soll.zustand + " (Güte " + soll.guete + ")");
          }
          else
          {
            _debug.finer("Übernahme Störfall-Zustand für " + _davProxy + "/" +
                verfahren + ": " + soll.zustand + " (Güte " + soll.guete + ")");
          }
          fireStoerfallListener();
        }
        else // es gibt bereits einen Wert für dieses Verfahren => glätten
        {
          long verzoegerung = getParameter().getRDSVerzoegerungsMatrix(ist.zustand, soll.zustand);

          if ((getConnection().getTime() - ist.getZeitStempel()) >= (verzoegerung * 1000))
          {
            // Verzoegerung für vorhanden Wert ist abgelaufen, übernehme neuen Wert sofort
            _aktuelleDaten.put(verfahren, soll);
            _debug.finer("Störfall-Zustand nach Glättung (" + verzoegerung + "s)" + "für " +
                _davProxy + "/" + verfahren + ": " + soll.zustand + " (Güte " + soll.guete + ")");
            fireStoerfallListener();
          }
          else
          {
            // Verzoegerung für vorhanden Wert ist noch nicht abgelaufen

            // Derselbe Zustand kommt mehrfach hintereinander
            if (ist.zustand.compareTo(soll.zustand) == 0)
            {
              // Wenn beim bestehenden Wert noch keine Verschlechterung aufgetreten ist,
              // wird der neue Wert übernommen, einschl. Zeitstempel, Güte
              if (!ist.getVerschlechtertung())
              {
                _aktuelleDaten.put(verfahren, soll);

                // Wenn sich die Güte geändert hat, muss  - ohne Glättung - upgedated werden
                if (ist.guete != soll.guete)
                {
                  _debug.finer("Störfall-Zustand nach Glättung (" + verzoegerung + "s)" + "für " +
                      _davProxy + "/" + verfahren + ": " + soll.zustand + " (Güte " + soll.guete + ")");

                  fireStoerfallListener();
                }
              }
              else
              {
                // Zurücksetzen Verschlechterung, da diese nur zwischen den letzten beiden
                // Eckwerten gültig sein soll
                ist.setVerschlechtertung(false);
                result = false;
              }
            }
            else // neuer Störfall innerhalb der Verzoegerungszeit wird nich übernommen
            {
              result = false;
            }
          }
        }
      }
      else // z. Z. nichtbenutzte Störfallverfahren
      {
        // diese Daten werden unmittelbar vorgehalten
        _aktuelleDaten.put(verfahren, soll);
      }
    }
    return result;
  }

  private class Daten
  {
    public Daten(Aspect aspekt, StoerfallZustand zustand, float guete, long zeitStempel)
    {
      this.aspekt = aspekt;
      this.zustand = zustand;
      this.guete = guete;
      _zeitStempel = zeitStempel;
    }

    public Daten(Daten daten)
    {
      this(daten.aspekt, daten.zustand, daten.guete, daten.getZeitStempel());
    }

    public final Aspect aspekt;

    public final StoerfallZustand zustand;

    public final float guete;

    private long _zeitStempel;

    public void setZeitStempel(long zeitStempel)
    {
      _zeitStempel = zeitStempel;
    }

    public long getZeitStempel()
    {
      return _zeitStempel;
    }

    private boolean _verschlechtertung;

    public void setVerschlechtertung(boolean verschlechtertung)
    {
      _verschlechtertung = verschlechtertung;
    }

    public boolean getVerschlechtertung()
    {
      return _verschlechtertung;
    }

    public Stoerfall getStoerfall()
    {
      return Stoerfall.this;
    }
  }

  private static class GlaettungsThread extends Thread
  {
    public GlaettungsThread()
    {
      super();
      setName("Störfall-Glättungs-Thread");
      start();
      _debug.config(this + " gestartet");
    }

    private Map<String, Daten> _queue = new LinkedHashMap<String, Daten>();

    public void put(Daten soll)
    {
       String key = soll.getStoerfall().getDavObjekt().getPid() + "/" + soll.aspekt.getPid();
       synchronized (_queue)
       {
         _queue.put(key, soll);
         _queue.notify();
       }
     }

    public Daten take()
       throws InterruptedException
     {
      synchronized (_queue)
      {
        while (_queue.isEmpty())
        {
          _queue.wait();
        }
        Iterator<Daten> it = _queue.values().iterator();
        Daten result = it.next();
        it.remove();
        return result;
      }
    }

    public final static long ZYKLUS = 200;

    @Override
    public void run()
    {
      int wartend = 0;

      while(true)
      {
        try
        {
          Daten daten = take();

          if (!daten.getStoerfall().uebernehmeDaten(daten))
          {
            put(daten);
            wartend += 1;
          }

          synchronized (_queue)
          {
            // Falls noch Einträge in der Queue warten, wird die
            // Queue spätestens nach der Zykluszeit wieder geprüft
            if (wartend > 0 && _queue.size() == wartend)
            {
              _queue.wait(ZYKLUS);
              wartend = 0; 
            }
          }
        }
        catch (InterruptedException e)
        {
          _debug.warning("Thread wurde unerwartet unterbrochen", e);
        }
      }  
    }
  }
   
}
